@@ -1,4 +1,4 @@
-# Copyright (C) 2011,2012 Eric Leblond <eric@regit.org>
+# Copyright (C) 2011-2018 Eric Leblond <eric@regit.org>
 #
 # You can copy, redistribute or modify this Program under the terms of
 # the GNU General Public License version 3 as published by the Free
@@ -14,16 +14,20 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 # 02110-1301, USA.
 
-from ConfigParser import SafeConfigParser
+try:
+    from configparser import SafeConfigParser
+except Exception:
+    from ConfigParser import SafeConfigParser
 from os import unlink, path, listdir, getcwd
 from string import Template
 from subprocess import Popen, PIPE, STDOUT
-from sys import stderr
+from sys import stderr, stdout
 from tempfile import NamedTemporaryFile
 import errno
 import re
+import sys
 
-COCCIGREP_VERSION = "1.13"
+COCCIGREP_VERSION = "1.20"
 
 have_multiprocessing = True
 try:
@@ -115,14 +119,15 @@ class CocciMatch:
 
     ptype_regexp = re.compile("^[ )]*\.")
 
-    def __init__(self, mfile, mline, mcol, mlineend, mcolend):
+    def __init__(self, mfile, mline, mcol, mlineend, mcolend, search):
         self.file = mfile
         self.line = int(mline)
         self.column = int(mcol)
         self.lineend = int(mlineend)
         self.columnend = int(mcolend)
+        self.search = search
         self.start_at = self.line
-        self.stop_at = self.line
+        self.stop_at = int(mlineend)
         self.trailer = ""
 
     def display(self, stype, mode='raw', oformat='term'):
@@ -141,7 +146,7 @@ class CocciMatch:
         f = open(self.file, 'r')
         lines = f.readlines()
         pmatch = lines[self.line - 1][self.column:self.columnend]
-        ptype = "*" # match is a pointer to struct
+        ptype = "*"  # match is a pointer to struct
         if (CocciMatch.ptype_regexp.search(lines[self.line - 1][self.columnend:])):
             ptype = ""
         output = ""
@@ -158,9 +163,18 @@ class CocciMatch:
             elif mode == 'emacs':
                 output += "%s:%s: (%s %s%s): %s" % (self.file, i + 1,
                 stype, ptype, pmatch, lines[i])
+            elif mode == 'grep' and stdout.isatty():
+                lineend = lines[i][self.columnend:]
+                if self.search.attribute:
+                    lineend = lineend.replace(self.search.attribute, "\033[0;31m" + self.search.attribute + "\033[0m", 1)
+                content = lines[i][:self.column] + \
+                    "\033[0;32m" + lines[i][self.column:self.columnend] + "\033[0m" \
+                    + lineend
+                output += "%s:%s:\t%s" % (self.file, i + 1, content)
             elif i == self.line - 1:
-                output += "%s:%s (%s %s%s): %s" % (self.file, i + 1,
-                stype, ptype, pmatch, lines[i])
+                content = lines[i]
+                output += "%s:%s (%s %s%s):\t%s" % (self.file, i + 1,
+                    stype, ptype, pmatch, content)
             else:
                 output += "%s-%s %s - %s" % (self.file, i + 1,
                 ' ' * (2 + len(stype + ptype + pmatch)), lines[i])
@@ -199,9 +213,9 @@ class CocciProcess:
             else:
                 output = Popen(self.cmd, stdout=PIPE,
                     stderr=PIPE).communicate()[0]
-        except Exception, err:
-            import pickle
-            output = pickle.dumps(err)
+        except Exception as err:
+            import json
+            output = json.dumps(err)
 
         self.input.send(output)
         self.input.close()
@@ -227,6 +241,7 @@ def _raise_run_err(err, cmd):
     raise CocciRunException("Unable to run '%s': %s." % (" ".join(cmd),
         err.strerror))
 
+
 class CocciPatch:
     """
     Class used to store information about a patch.
@@ -256,13 +271,13 @@ class CocciPatch:
 
     def __getitem__(self, key):
         # name ok?
-        if not key in CocciPatch.keywords:
+        if key not in CocciPatch.keywords:
             raise KeyError("Trying to get invalid name '%s'." % key)
         return self.__dict__[key]
 
     def __setitem__(self, key, value):
         # name ok?
-        if not key in CocciPatch.keywords:
+        if key not in CocciPatch.keywords:
             raise KeyError("Trying to set invalid name '%s'." % key)
         # set
         self.__dict__[key] = value
@@ -275,9 +290,10 @@ class CocciPatch:
             return out + ": No info available\n"
         for key in self:
             if key not in ["Name", "Desc", "File"]:
-                out +=" * %s: %s\n" % (key, self[key])
+                out += " * %s: %s\n" % (key, self[key])
         return out
 
+
 class CocciGrep:
     """
     Core class of the module: setup and run.
@@ -286,14 +302,20 @@ class CocciGrep:
     of initialisation and running of the request.
     """
     spatch = "spatch"
+    cocci_python_hdr_std = """
+@ script:python @
+"""
+    cocci_python_hdr_filter = """
+@ script:python depends on filter @
+"""
+
     cocci_python = """
 
-@ script:python @
 p1 << init.p1;
 @@
 
 for p in p1:
-    print "%s:%s:%s:%s:%s" % (p.file,p.line,p.column,p.line_end,p.column_end)
+    print("%s:%s:%s:%s:%s" % (p.file,p.line,p.column,p.line_end,p.column_end))
 """
 
     def __init__(self):
@@ -320,9 +342,6 @@ for p in p1:
         :type operation: str
         :raise: :class:`CocciRunException`
         """
-        if stype == None:
-            raise CocciRunException("Can't use coccigrep without type to "
-                "search")
         self.type = stype
         self.attribute = attribute
         self.operation = operation
@@ -374,7 +393,7 @@ for p in p1:
 
         :return: list of operations in a list of str
         """
-        return self.operations.keys()
+        return list(self.operations.keys())
 
     def get_operation_name(self, fname):
         return _operation_name(fname)
@@ -405,13 +424,13 @@ for p in p1:
         self.verbose = True
 
     def get_spatch_version(self):
-        cmd = [self.spatch] + [ '-version']
+        cmd = [self.spatch] + ['-version']
         try:
             output = Popen(cmd, stdout=PIPE, stderr=STDOUT).communicate()[0]
-        except OSError, err:
+        except OSError as err:
             _raise_run_err(err, cmd)
         reg = r"version (.*?) with"
-        m = re.search(reg, output)
+        m = re.search(reg, output.decode('utf8'))
         return m.group(1)
 
     def spatch_newer_than(self, version):
@@ -454,13 +473,19 @@ for p in p1:
         # do substitution
         cocci_smpl = cocci_smpl_tmpl.substitute(type=self.type,
             attribute=self.attribute, cocci_regexp_equal=cocci_op)
-        cocci_grep = cocci_smpl + CocciGrep.cocci_python
+        if '@filter@' in cocci_smpl:
+            cocci_grep = cocci_smpl + CocciGrep.cocci_python_hdr_filter + CocciGrep.cocci_python
+        else:
+            cocci_grep = cocci_smpl + CocciGrep.cocci_python_hdr_std + CocciGrep.cocci_python
 
-        tmp_cocci_file.write(cocci_grep)
+        if sys.version < '3':
+            tmp_cocci_file.write(cocci_grep)
+        else:
+            tmp_cocci_file.write(bytes(cocci_grep, 'UTF-8'))
         tmp_cocci_file.flush()
 
         # launch spatch
-        output = ""
+        output = "".encode('utf8')
         # Launch parallel spatch
         if self.ncpus > 1 and len(files) > 1:
             fseq = []
@@ -475,20 +500,25 @@ for p in p1:
                 cmd = [self.spatch]
                 cmd += self.options
                 cmd += ["-sp_file", tmp_cocci_file.name]
+                for cfile in sub_files:
+                    include_dir = path.dirname(cfile)
+                    if len(include_dir):
+                        cmd += ["-I", include_dir]
+                        break
                 cmd += sub_files
                 sprocess = CocciProcess(cmd, self.verbose)
                 sprocess.start()
                 self.process.append(sprocess)
             for process in self.process:
-                ret = process.recv()
+                ret = process.recv().decode('utf8')
                 process.join()
                 if not ret.startswith('cexceptions\n'):
                     # CocciProcess return a serialized exception
                     # in case of exception
-                    output += ret
+                    output += ret.encode('utf8')
                     continue
-                import pickle
-                err = pickle.loads(ret)
+                import json
+                err = json.loads(ret)
                 _raise_run_err(err, cmd)
             tmp_cocci_file.close()
         # Fallback to one spatch
@@ -497,30 +527,26 @@ for p in p1:
             cmd += self.options
             cmd += ["-sp_file", tmp_cocci_file.name]
             cmd += files
+            include_dir = path.dirname(files[0])
+            if len(include_dir):
+                cmd += ["-I", include_dir]
             try:
                 if self.verbose:
                     stderr.write("Running: %s.\n" % " ".join(cmd))
                     output = Popen(cmd, stdout=PIPE).communicate()[0]
                 else:
                     output = Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()[0]
-            except OSError, err:
+            except OSError as err:
                 unlink(tmp_cocci_file_name)
                 _raise_run_err(err, cmd)
 
             tmp_cocci_file.close()
 
-        prevfile = None
-        prevline = None
         self.matches = []
-        for ematch in output.split("\n"):
+        for ematch in output.decode('utf8').split("\n"):
             try:
                 (efile, eline, ecol, elinend, ecolend) = ematch.split(":")
-                nmatch = CocciMatch(efile, eline, ecol, elinend, ecolend)
-                # if there is equality then we will already display the line
-                if (efile == prevfile) and (eline == prevline):
-                    continue
-                prevfile = efile
-                prevline = eline
+                nmatch = CocciMatch(efile, eline, ecol, elinend, ecolend, self)
                 self.matches.append(nmatch)
             except ValueError:
                 pass
@@ -541,7 +567,7 @@ for p in p1:
         """
         if before != 0 or after != 0:
             prev_match = None
-            for index in xrange(len(self.matches)):
+            for index in range(len(self.matches)):
                 cur_match = self.matches[index]
                 cur_match.start_at = cur_match.line - before
                 cur_match.stop_at = cur_match.line + after
@@ -557,7 +583,7 @@ for p in p1:
                             cur_match.start_at = prev_match.stop_at + 1
 
                         if prev_match.stop_at + 1 == cur_match.start_at:
-                            # No seprator if groups are contiguous
+                            # No separator if groups are contiguous
                             prev_match.trailer = ""
 
                 prev_match = cur_match
